@@ -53,9 +53,10 @@ def accuracy_metrics(vals, expected_dt, truth_val):
     }
 
 
-# def completeness(row):
-#     vals = row[row.notna()]
-#     return len(vals) / len(row)
+def completeness_metric(vals, expected_dt, truth_val):
+    expected_vals = vals[expected_dt:]
+    not_na = expected_vals[expected_vals.notna()]
+    return {"Completeness": len(not_na) / len(not_na)}
 
 
 def time_till_metrics(vals, expected_dt, truth_val):
@@ -114,6 +115,7 @@ def validity_nondecreasing_smoothness(series, *args, **kwargs):
     orig_diffs = series.diff()
     rises = [abs(diff) for diff in orig_diffs if diff > 0]
     inds = [ind for ind, diff in enumerate(orig_diffs) if diff > 0]
+
     if len(inds) == 0:
         return 0
 
@@ -125,7 +127,7 @@ def validity_nondecreasing_smoothness(series, *args, **kwargs):
 def validity_nondecreasing(series, *args, **kwargs):
     """Checks if values are nondecreasing.
     Return number of times a decrease occurs"""
-    return len(series[series.diff() < 0])
+    return len(series[series.notna().diff() < 0])
 
 
 # ----------------------------------------------#
@@ -165,6 +167,7 @@ def apply_to_updates(func):
     :param func:
     :return: func
     """
+
     def over_updates(data_df, truth_df, expected_df, *args, **kwargs):
         dct = {}
 
@@ -172,7 +175,7 @@ def apply_to_updates(func):
 
             not_na = series[series.notna()]
             if len(not_na) == 0:
-                dct[update] = {}
+                dct[update] = None
 
             else:
                 dct[update] = func(series, truth_df, expected_df, *args, **kwargs)
@@ -182,7 +185,6 @@ def apply_to_updates(func):
     return over_updates
 
 
-
 def timeliness(data_df, truth_df, expected_df):
     """
 
@@ -222,17 +224,28 @@ def validity(data_df, truth_df, expected_df):
     return func(data_df, truth_df, expected_df)
 
 
-#
-#
-# def consistency(data_df, truth_df, expected_df):
-#     func = apply_to_observations(accuracy_metrics)
-#     return func(data_df, truth_df, expected_df)
-#
-#
-# def completeness(data_df, truth_df, expected_df):
-#     func1 = apply_to_observations(accuracy_metrics)
-#     func2 = apply_to_updates(accuracy_metrics)
-#     return func(data_df, truth_df, expected_df)
+def consistency(data_df, truth_df, expected_df):
+    updates = data_df.columns
+    dct = {}
+    last_non_null = None
+    for prior, post in zip(updates[:-1], updates[1:]):
+
+        if len(data_df[prior].notna()) > 0:
+            last_non_null = data_df[prior]
+        post_nonnulls = data_df[post].notna()
+        diff = data_df[post][post_nonnulls] - last_non_null[post_nonnulls]
+
+        if len(diff) > 0:
+            dct[post] = {"ConsistencyPercentUpdated": len(diff[diff > 0]) / len(diff)}
+        else:
+            dct[post] = None
+
+    return pd.DataFrame.from_dict(dct)
+
+
+def completeness(data_df, truth_df, expected_df):
+    func = apply_to_observations(completeness_metric)
+    return func(data_df, truth_df, expected_df)
 
 
 # ----------------------------------------------#
@@ -248,8 +261,9 @@ def df_from_s3_csv(bucket, key):
     :return: pandas dataframe
     """
     s3_client = boto3.client('s3')
-
+    logging.info(f'Attempting to read S3 object {key}')
     obj = s3_client.get_object(Bucket=bucket, Key=key)
+    logging.info(f'Successfully read S3 object {key}')
     df = pd.read_csv(obj['Body'], sep=',', index_col=0)
     return df
 
@@ -324,7 +338,9 @@ def metric_node(
     funcs_mapping = {
         "accuracy": accuracy,
         "timeliness": timeliness,
-        "validity": validity
+        "validity": validity,
+        "consistency": consistency,
+        "completeness": completeness
     }
 
     logging.basicConfig(level=logging.DEBUG,
@@ -387,3 +403,108 @@ def metric_node(
         'statusCode': status,
         'body': message
     }
+
+
+def summary_node(
+        event,
+        context,
+        to_sqs_url,
+        region_name,
+        to_bucket,
+        to_key
+):
+    """
+     Summarize the Data Quality Dimensions
+
+    Load a pandas dataframe from the buck/key referenced in the triggering event
+    Execute func on the dataframe which must return a df
+    Put the df into to_bucket/key
+    Send message to to_sqs_url
+
+
+    """
+    logging.basicConfig(level=logging.DEBUG,
+                        format='%(levelname)s: %(asctime)s: %(message)s')
+    dims_files = [
+        "timeliness.csv",
+        "accuracy.csv",
+        "validity.csv",
+        "completeness.csv",
+        "consistency.csv"
+    ]
+
+    body = event["Records"][0]["body"]
+    if isinstance(body, str):
+        body = json.loads(body)
+
+    path, data_file, truth_file, expected_file = "", "data.csv", "truth.csv", "expected_update.csv"
+    if 'path' in body:
+        path = body['path']
+    if 'data_file' in body:
+        data_file = body['data_file']
+    if 'truth_file' in body:
+        truth_file = body['truth_file']
+    if 'expected_file' in body:
+        expected_file = body['expected_file']
+
+    from_bucket = body['bucket']
+
+    def make_key(file):
+        if path == "":
+            return file
+        else:
+            return "/".join([path, file])
+
+    completed_stations = [each for each in body['completed_stations']]
+
+    dfs = []
+    transpose = ["validity.csv", "consistency.csv"]
+
+    for from_key in dims_files:
+        df = df_from_s3_csv(from_bucket, make_key(from_key))
+
+        if from_key in transpose:
+            df = df.T
+
+        desc_cols = ['mean', '50%', 'std']
+        df_desc = df.describe()
+        df_desc = df_desc.loc[desc_cols]
+        df_desc.index = ['mean', 'median', 'std']
+        df_desc = df_desc.T
+        df_desc["coef_var"] = df_desc['std'] / df_desc['mean']
+        kurtosis = df.kurtosis()
+        df_desc = pd.concat([df_desc, kurtosis], axis=1)
+        df_desc.rename({0: "kurtosis"}, inplace=True, axis=1)
+        dfs.append(df_desc)
+
+    finished_df = pd.concat(dfs)
+
+    status = df_to_s3_csv(bucket=to_bucket, key=make_key(to_key), df=finished_df)
+
+    if status == 200:
+        logging.info(f'Successful S3 put_object response')
+
+    else:
+        logging.info(f'Unsuccessful S3 put_object response - {status}')
+
+    completed_stations.append(context.function_name)
+
+    message = {
+        'bucket': from_bucket,
+        'path': path,
+        'data_file': data_file,
+        'truth_file': truth_file,
+        'expected_file': expected_file,
+        'summary_file': to_key,
+        'completed_stations': completed_stations
+    }
+
+    msg = send_sqs_message(to_sqs_url, message, region_name)
+
+    if msg is not None:
+        logging.info(f'Sent SQS message ID: {msg["MessageId"]}')
+
+    return {
+        'statusCode': status,
+        'body': message
+    }
